# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import os
import logging
import random
import warnings

import numpy as np
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.multiprocessing as mp

from semilearn.algorithms import get_algorithm, name2alg
from semilearn.imb_algorithms import get_imb_algorithm, name2imbalg
from semilearn.core.utils import get_net_builder, get_logger, get_port, send_model_cuda, count_parameters, \
    over_write_args_from_file, TBLog
from torch.utils.data import DataLoader
from semilearn.core.utils import get_net_builder, get_dataset
from sklearn.metrics import accuracy_score, balanced_accuracy_score, precision_score, recall_score, f1_score, confusion_matrix
from semilearn.core.utils import get_dataset, get_data_loader, send_model_cuda, IMB_ALGORITHMS

def get_config():
    from semilearn.algorithms.utils import str2bool

    parser = argparse.ArgumentParser(description='Semi-Supervised Learning (USB)')

    '''
    Saving & loading of the model.
    '''
    parser.add_argument('--save_dir', type=str, default='./saved_models')
    parser.add_argument('-sn', '--save_name', type=str, default='fixmatch')
    parser.add_argument('--resume', action='store_true')
    parser.add_argument('--load_path', type=str)
    parser.add_argument('-o', '--overwrite', action='store_true', default=True)
    parser.add_argument('--use_tensorboard', action='store_true', help='Use tensorboard to plot and save curves')
    parser.add_argument('--use_wandb', action='store_true', help='Use wandb to plot and save curves')
    parser.add_argument('--use_aim', action='store_true', help='Use aim to plot and save curves')

    '''
    Training Configuration of FixMatch
    '''
    parser.add_argument('--epoch', type=int, default=1)
    parser.add_argument('--num_train_iter', type=int, default=20,
                        help='total number of training iterations')
    parser.add_argument('--num_warmup_iter', type=int, default=0,
                        help='cosine linear warmup iterations')
    parser.add_argument('--num_eval_iter', type=int, default=10,
                        help='evaluation frequency')
    parser.add_argument('--num_log_iter', type=int, default=5,
                        help='logging frequencu')
    parser.add_argument('-nl', '--num_labels', type=int, default=400)
    parser.add_argument('-bsz', '--batch_size', type=int, default=8)
    parser.add_argument('--uratio', type=int, default=1,
                        help='the ratio of unlabeled data to labeld data in each mini-batch')
    parser.add_argument('--eval_batch_size', type=int, default=16,
                        help='batch size of evaluation data loader (it does not affect the accuracy)')
    parser.add_argument('--ema_m', type=float, default=0.999, help='ema momentum for eval_model')
    parser.add_argument('--ulb_loss_ratio', type=float, default=1.0)

    '''
    Optimizer configurations
    '''
    parser.add_argument('--optim', type=str, default='SGD')
    parser.add_argument('--lr', type=float, default=3e-2)
    parser.add_argument('--momentum', type=float, default=0.9)
    parser.add_argument('--weight_decay', type=float, default=5e-4)
    parser.add_argument('--layer_decay', type=float, default=1.0,
                        help='layer-wise learning rate decay, default to 1.0 which means no layer decay')

    '''
    Backbone Net Configurations
    '''
    parser.add_argument('--net', type=str, default='wrn_28_2')
    parser.add_argument('--net_from_name', type=str2bool, default=False)
    parser.add_argument('--use_pretrain', default=False, type=str2bool)
    parser.add_argument('--pretrain_path', default='', type=str)

    '''
    Algorithms Configurations
    '''

    ## core algorithm setting
    parser.add_argument('-alg', '--algorithm', type=str, default='fixmatch', help='ssl algorithm')
    parser.add_argument('--use_cat', type=str2bool, default=True, help='use cat operation in algorithms')
    parser.add_argument('--amp', type=str2bool, default=False, help='use mixed precision training or not')
    parser.add_argument('--clip_grad', type=float, default=0)

    ## imbalance algorithm setting
    parser.add_argument('-imb_alg', '--imb_algorithm', type=str, default=None, help='imbalance ssl algorithm')

    '''
    Data Configurations
    '''

    ## standard setting configurations
    parser.add_argument('--data_dir', type=str, default='./data')
    parser.add_argument('-ds', '--dataset', type=str, default='cifar10')
    parser.add_argument('-nc', '--num_classes', type=int, default=10)
    parser.add_argument('--train_sampler', type=str, default='RandomSampler')
    parser.add_argument('--num_workers', type=int, default=1)
    parser.add_argument('--include_lb_to_ulb', type=str2bool, default='True',
                        help='flag of including labeled data into unlabeled data, default to True')

    ## imbalanced setting arguments
    parser.add_argument('--lb_imb_ratio', type=int, default=1, help="imbalance ratio of labeled data, default to 1")
    parser.add_argument('--ulb_imb_ratio', type=int, default=1, help="imbalance ratio of unlabeled data, default to 1")
    parser.add_argument('--ulb_num_labels', type=int, default=None,
                        help="number of labels for unlabeled data, used for determining the maximum number of labels in imbalanced setting")

    ## cv dataset arguments
    parser.add_argument('--img_size', type=int, default=32)
    parser.add_argument('--crop_ratio', type=float, default=0.875)

    ## nlp dataset arguments
    parser.add_argument('--max_length', type=int, default=512)

    ## speech dataset algorithms
    parser.add_argument('--max_length_seconds', type=float, default=4.0)
    parser.add_argument('--sample_rate', type=int, default=16000)

    '''
    multi-GPUs & Distrbitued Training
    '''

    ## args for distributed training (from https://github.com/pytorch/examples/blob/master/imagenet/main.py)
    parser.add_argument('--world-size', default=1, type=int,
                        help='number of nodes for distributed training')
    parser.add_argument('--rank', default=0, type=int,
                        help='**node rank** for distributed training')
    parser.add_argument('-du', '--dist-url', default='tcp://127.0.0.1:11111', type=str,
                        help='url used to set up distributed training')
    parser.add_argument('--dist-backend', default='nccl', type=str,
                        help='distributed backend')
    parser.add_argument('--seed', default=1, type=int,
                        help='seed for initializing training. ')
    parser.add_argument('--gpu', default=None, type=int,
                        help='GPU id to use.')
    parser.add_argument('--multiprocessing-distributed', type=str2bool, default=False,
                        help='Use multi-processing distributed training to launch '
                             'N processes per node, which has N GPUs. This is the '
                             'fastest way to use PyTorch for either single node or '
                             'multi node data parallel training')
    # config file
    parser.add_argument('--c', type=str, default='')

    # add algorithm specific parameters
    args = parser.parse_args()
    over_write_args_from_file(args, args.c)
    for argument in name2alg[args.algorithm].get_argument():
        parser.add_argument(argument.name, type=argument.type, default=argument.default, help=argument.help)

    # add imbalanced algorithm specific parameters
    args = parser.parse_args()
    over_write_args_from_file(args, args.c)
    if args.imb_algorithm is not None:
        for argument in name2imbalg[args.imb_algorithm].get_argument():
            parser.add_argument(argument.name, type=argument.type, default=argument.default, help=argument.help)
    args = parser.parse_args()
    over_write_args_from_file(args, args.c)
    return args


def main(args):
    '''
    For (Distributed)DataParallelism,
    main(args) spawn each process (main_worker) to each GPU.
    '''

    assert args.num_train_iter % args.epoch == 0, \
        f"# total training iter. {args.num_train_iter} is not divisible by # epochs {args.epoch}"

    save_path = os.path.join(args.save_dir, args.save_name)
    if os.path.exists(save_path) and args.overwrite and args.resume == False:
        import shutil
        shutil.rmtree(save_path)
    if os.path.exists(save_path) and not args.overwrite:
        raise Exception('already existing model: {}'.format(save_path))
    if args.resume:
        if args.load_path is None:
            raise Exception('Resume of training requires --load_path in the args')
        if os.path.abspath(save_path) == os.path.abspath(args.load_path) and not args.overwrite:
            raise Exception('Saving & Loading pathes are same. \
                            If you want over-write, give --overwrite in the argument.')

    if args.seed is not None:
        warnings.warn('You have chosen to seed training. '
                      'This will turn on the CUDNN deterministic setting, '
                      'which can slow down your training considerably! '
                      'You may see unexpected behavior when restarting '
                      'from checkpoints.')

    if args.gpu == 'None':
        args.gpu = None
    if args.gpu is not None:
        warnings.warn('You have chosen a specific GPU. This will completely '
                      'disable data parallelism.')

    if args.dist_url == "env://" and args.world_size == -1:
        args.world_size = int(os.environ["WORLD_SIZE"])

    # distributed: true if manually selected or if world_size > 1
    args.distributed = args.world_size > 1 or args.multiprocessing_distributed
    ngpus_per_node = torch.cuda.device_count()  # number of gpus of each node

    if args.multiprocessing_distributed:
        # now, args.world_size means num of total processes in all nodes
        args.world_size = ngpus_per_node * args.world_size

        # args=(,) means the arguments of main_worker
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args))
    else:
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args):
    '''
    main_worker is conducted on each GPU.
    '''

    global best_acc1
    args.gpu = gpu

    # random seed has to be set for the syncronization of labeled data sampling in each process.
    assert args.seed is not None
    random.seed(args.seed)
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)
    cudnn.deterministic = True
    cudnn.benchmark = True

    # SET UP FOR DISTRIBUTED TRAINING
    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])

        if args.multiprocessing_distributed:
            args.rank = args.rank * ngpus_per_node + gpu  # compute global rank

        # set distributed group:
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)

    # SET save_path and logger
    save_path = os.path.join(args.save_dir, args.save_name)
    logger_level = "WARNING"
    tb_log = None
    if args.rank % ngpus_per_node == 0:
        tb_log = TBLog(save_path, 'tensorboard', use_tensorboard=args.use_tensorboard)
        logger_level = "INFO"

    logger = get_logger(args.save_name, save_path, logger_level)
    logger.info(f"Use GPU: {args.gpu} for training")

    _net_builder = get_net_builder(args.net, args.net_from_name)
    # optimizer, scheduler, datasets, dataloaders with be set in algorithms
    if args.imb_algorithm is not None:
        model = get_imb_algorithm(args, _net_builder, tb_log, logger)
    else:
        model = get_algorithm(args, _net_builder, tb_log, logger)
    logger.info(f'Number of Trainable Params: {count_parameters(model.model)}')

    # SET Devices for (Distributed) DataParallel
    model.model = send_model_cuda(args, model.model)
    model.ema_model = send_model_cuda(args, model.ema_model, clip_batch=False)
    # logger.info(f"Arguments: {model.args}")
    # model.load_model(args.load_path)
    checkpoint_path = os.path.join(args.load_path)
    checkpoint = torch.load(checkpoint_path)
    #load_model = checkpoint['ema_model']
    load_model = checkpoint['model']
    load_state_dict = {}

    # for param_name in model.model.state_dict():
    #     print(param_name)
    # print('hoho')

    for key, item in load_model.items():
        if key.startswith('module'):
            new_key = '.'.join(key.split('.')[1:])
            load_state_dict[new_key] = item
        else:
            load_state_dict[key] = item
    model.model.module.load_state_dict(load_state_dict)#.module

    # If args.resume, load checkpoints from args.load_path
    # if args.resume and os.path.exists(args.load_path):
    #     try:
    #         model.load_model(args.load_path)
    #     except:
    #         logger.info("Fail to resume load path {}".format(args.load_path))
    #         args.resume = False

    # START TRAINING of FixMatch
    logger.info("Model testing")

    args.num_labels = 40
    args.ulb_num_labels = 10#49600
    args.lb_imb_ratio = 1
    args.ulb_imb_ratio = 1
    args.seed = 0
    args.epoch = 1
    args.num_train_iter = 1024
    dataset_dict = get_dataset(args, 'fixmatch', args.dataset, args.num_labels, args.num_classes, args.data_dir, False)
    #dataset_dict = get_dataset(args, 'fixmatch', args.dataset, 60000, args.num_classes, args.data_dir, False)
    # eval_dset = dataset_dict['eval']
    # eval_loader = DataLoader(eval_dset, batch_size=args.batch_size, drop_last=False, shuffle=False, num_workers=4)
    # eval_loader=model.loader_dict['train_ulb']#eval,train_lb
    # for param_name in load_state_dict:
    #     print(param_name)

    eval_loader = get_data_loader(args,
                                  model.dataset_dict['train_lb'],#'train_ulb'
                                  100,#args.batch_size,
                                  data_sampler=args.train_sampler,
                                  drop_last=False,
                                  num_iters=100,#num_iters=args.num_train_iter,
                                  num_epochs=1,#args.num_workers
                                  num_workers=args.num_workers,
                                  distributed=args.distributed)


    acc = 0.0
    test_feats = []
    test_preds = []
    test_probs = []
    test_labels = []

    saved_features = None
    saved_ids = None
    saved_preds= None
    saved_mask= None
    model.model.eval()

    with torch.no_grad():
        aa=0
        for data in eval_loader:
            #有标签数据
            # image = data['x_lb']
            # target = data['y_lb']
            #无标签数据
            print(data.keys())
            ##image = data['x_ulb_w']
            #image = torch.tensor(model.dataset_dict['train_ulb'].data[data['idx_ulb']])
            ##target = model.dataset_dict['train_ulb'].targets[data['idx_ulb']]#'train_ulb'
            
            image = data['x_lb']
            
            target = model.dataset_dict['train_lb'].targets[data['idx_lb']]
            
            #target = data['idx_ulb']
            target = torch.from_numpy(target)

            image = image.type(torch.FloatTensor).cuda()
            # feat = model.model.backbone(image, only_feat=True)
            # logit = model.model.teacher_classifier(feat, only_fc=True)
            # feat = model.model.backbone(image)['feat']
            # logit = model.model.aux_classifier(feat)
            #feat = model.model.module.backbone(image)['feat']
            #logit = model.model.module.backbone(image)['logits']

            # feat = model.model(image, only_feat=True)
            # logit = model.model(image)['logits']
            # prob = logit.softmax(dim=-1)
            # pred = prob.argmax(1)


            # feat = model.model.module.backbone(image, only_feat=True)# 当用其他分类器时
            # feat = model.model.module(image, only_feat=True) #使用春fix
            #feat = model.model.module(image)['feat']
            feat = model.model.module(image)['feat']
            logit = model.model.module(image)['logits']

            max_probs, pred = torch.max(model.compute_prob(logit.detach()), dim=-1)
            feat = model.compute_prob(logit.detach())

            #about ulb

            mask_ulb_1 = max_probs.ge(0.95).to(feat.dtype)
            if saved_mask is not None:
                saved_mask = torch.cat((saved_mask, mask_ulb_1), 0)
            else:
                saved_mask = mask_ulb_1


            # prob = logit.softmax(dim=-1)
            # pred = prob.argmax(1)

            acc += pred.cpu().eq(target).numpy().sum()

            if saved_features is not None:
                saved_features = torch.cat((saved_features, feat), 0)
            else:
                saved_features = feat
            if saved_ids is not None:
                saved_ids = torch.cat((saved_ids, target), 0)
            else:
                saved_ids = target
            if saved_preds is not None:
                saved_preds = torch.cat((saved_preds, pred), 0)
            else:
                saved_preds = pred

            test_feats.append(feat.cpu().numpy())
            test_preds.append(pred.cpu().numpy())
            # test_probs.append(prob.cpu().numpy())
            test_labels.append(target.cpu().numpy())
            aa+=1
    test_feats = np.concatenate(test_feats)
    test_preds = np.concatenate(test_preds)
    # test_probs = np.concatenate(test_probs)
    test_labels = np.concatenate(test_labels)

    cf_mat = confusion_matrix(test_labels, test_preds, normalize='true')

    #print(f"Test Accuracy: {acc / len(eval_dset)}")
    print(f"Test Accuracy: {accuracy_score(test_labels, test_preds)}")
    print(f"confusion_matrix:\n {cf_mat}")
    np.save('features_raw.npy', saved_features.detach().cpu().numpy())
    np.save('results_raw.npy', saved_ids.detach().cpu().numpy())
    # np.save('pred_raw.npy', saved_ids.detach().cpu().numpy())
    np.save('pred_raw.npy', saved_preds.detach().cpu().numpy())

    np.save('mask_raw.npy', saved_mask.detach().cpu().numpy())

    if hasattr(model, 'finetune'):
        logger.info("Finetune stage")
        model.finetune()

    logging.warning(f"GPU {args.rank} training is FINISHED")


if __name__ == "__main__":
    args = get_config()
    port = get_port()
    # args.net = get_net_builder(args.net, args.net_from_name)(num_classes=args.num_classes)
    args.dist_url = "tcp://127.0.0.1:" + str(port)
    main(args)
